"""
This file is part of mpunittest, a parallel unittest runner.
Copyright (C) 2023 Maximilian Pilz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import itertools
import multiprocessing
import typing
import unittest
import unittest.mock

import mpunittest.comm


class TestComm(unittest.TestCase):

    def test_shared_mem_buf(self):

        for i, j in itertools.product(range(1, 10), range(3, 12)):
            with unittest.mock.patch(target='mpunittest.comm._BYTES_PER_TRANSMISSION_CODE',
                                     new=i), \
                    unittest.mock.patch(target='mpunittest.comm._DURATION_SIZE',
                                        new=i * 7):

                sh_mem_comm = mpunittest.comm.SharedMemoryCommunicator(code_count=j,
                                                                       mp_context=multiprocessing.get_context('spawn'))
                try:
                    self.assertEqual(bytearray([0] * i * j),
                                     bytearray(sh_mem_comm._shmm_for_results.buf[0:i * j]))

                    self.assertEqual(bytearray([0] * i * 7 * j),
                                     bytearray(sh_mem_comm._shmm_for_durations.buf[0:i * 7 * j]))

                    sh_mem_comm.get()

                    self.assertEqual(bytearray([0] * (i - 1) + [7] + [0] * i * (j - 1)),
                                     bytearray(sh_mem_comm._shmm_for_results.buf[0:i * j]))

                    sh_mem_comm.get()

                    self.assertEqual(bytearray(([0] * (i - 1) + [7]) * 2 + [0] * i * (j - 2)),
                                     bytearray(sh_mem_comm._shmm_for_results.buf[0:i * j]))

                    sh_mem_comm.get()

                    self.assertEqual(bytearray(([0] * (i - 1) + [7]) * 3 + [0] * i * (j - 3)),
                                     bytearray(sh_mem_comm._shmm_for_results.buf[0:i * j]))

                    self.assertEqual(bytearray([0] * i * 7 * j),
                                     bytearray(sh_mem_comm._shmm_for_durations.buf[0:i * 7 * j]))

                    sh_mem_comm.put(1, mpunittest.comm.TransmissionCode.FAIL, 50)

                    self.assertEqual(bytearray([0] * i * 7 + [0] * (i * 7 - 1) + [50] + [0] * i * 7 * (j - 2)),
                                     bytearray(sh_mem_comm._shmm_for_durations.buf[0:i * 7 * j]))

                    self.assertEqual(
                        bytearray([0] * (i - 1) + [7] + [0] * (i - 1) + [30] + [0] * (i - 1) + [7] + [0] * i * (j - 3)),
                        bytearray(sh_mem_comm._shmm_for_results.buf[0:i * j]))

                    results: typing.List[
                        typing.Tuple[int, mpunittest.comm.TransmissionCode, int]] = sh_mem_comm.get_results()

                    self.assertEqual(
                        [(0, mpunittest.comm.TransmissionCode.STARTED, 0),
                         (1, mpunittest.comm.TransmissionCode.FAIL, 50),
                         (2, mpunittest.comm.TransmissionCode.STARTED, 0),
                         *((k + 3, mpunittest.comm.TransmissionCode.INITIAL, 0) for k in range(j - 3))
                         ],
                        results)

                    self.assertFalse(sh_mem_comm.is_finished())

                    for m in range(j - 3):
                        sh_mem_comm.get()
                        sh_mem_comm.put(index=m + 3,
                                        transmission_code=mpunittest.comm.TransmissionCode.PASS,
                                        duration=m)

                    results: typing.List[
                        typing.Tuple[int, mpunittest.comm.TransmissionCode, int]] = sh_mem_comm.get_results()

                    self.assertEqual(
                        [(0, mpunittest.comm.TransmissionCode.STARTED, 0),
                         (1, mpunittest.comm.TransmissionCode.FAIL, 50),
                         (2, mpunittest.comm.TransmissionCode.STARTED, 0),
                         *((k + 3, mpunittest.comm.TransmissionCode.PASS, k) for k in range(j - 3))
                         ],
                        results)

                    sh_mem_comm.put(index=0,
                                    transmission_code=mpunittest.comm.TransmissionCode.FAIL,
                                    duration=13)
                    sh_mem_comm.put(index=2,
                                    transmission_code=mpunittest.comm.TransmissionCode.FAIL,
                                    duration=17)

                    self.assertTrue(sh_mem_comm.is_finished())

                finally:
                    try:
                        sh_mem_comm.close()
                    finally:
                        sh_mem_comm.unlink()


if __name__ == '__main__':
    unittest.main()



"""
This file is part of mpunittest, a parallel unittest runner.
Copyright (C) 2023 Maximilian Pilz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import logging
import pathlib

import mpunittest.logging
import mpunittest.runner
import mpunittest.tests.dummy.dirs

if __name__ == '__main__':
    handler = logging.StreamHandler()

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)
    mpunittest.logging.logger.addHandler(handler)

    merging_runner = mpunittest.runner.MergingRunner(process_count=7)

    result = merging_runner.discover_and_run(
        start_dir=pathlib.Path(mpunittest.tests.dummy.dirs.__file__).parent.resolve(),
        pattern="*test.py",
        html_result_assets=mpunittest.runner.HtmlResultAssets(
            document_title='test1',
            document_file_name='test2',
            result_path=None  # this will cause usage of the default dir
        )
    )

    mpunittest.logging.logger.getChild('tests.real.manual').info(
        'run finished with return value %s', result)



"""
This file is part of mpunittest, a parallel unittest runner.
Copyright (C) 2023 Maximilian Pilz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import abc
import enum
import multiprocessing
import multiprocessing.shared_memory
import typing

_DURATION_SIZE = 64


class TransmissionCode(int, enum.Enum):
    INITIAL = 0b000000
    STARTED = 0b000111
    PASS = 0b011001
    FAIL = 0b011110
    SKIPPED = 0b101010
    ERROR = 0b101101


_BITS_OF_MAX_VALUE = 0xFF.bit_length()
assert _BITS_OF_MAX_VALUE >= 8


def _length_in_bytes(i: int) -> int:
    assert i >= 0

    if i == 0:
        return 1

    return (i.bit_length() + _BITS_OF_MAX_VALUE - 1) // _BITS_OF_MAX_VALUE


assert len(TransmissionCode) > 1
_BYTES_PER_TRANSMISSION_CODE = max(_length_in_bytes(e) for e in TransmissionCode)


class BaseCommunicator(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __init__(self, code_count: int, mp_context: multiprocessing.context.BaseContext):
        assert code_count > 0

        self._code_count: int = code_count
        self._mp_context: multiprocessing.context.BaseContext = mp_context

    @abc.abstractmethod
    def get(self):
        raise NotImplementedError

    @abc.abstractmethod
    def put(self, index, transmission_code, duration):
        raise NotImplementedError

    @abc.abstractmethod
    def get_results(self) -> typing.List[typing.Tuple[int, TransmissionCode, int]]:
        raise NotImplementedError

    @abc.abstractmethod
    def is_finished(self):
        raise NotImplementedError

    @abc.abstractmethod
    def close(self):
        raise NotImplementedError

    @abc.abstractmethod
    def unlink(self):
        raise NotImplementedError


class SharedMemoryCommunicator(BaseCommunicator):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._from_bytes_kwargs = {'byteorder': 'big',
                                   'signed': False}
        self._to_bytes_kwargs_without_length = self._from_bytes_kwargs
        self._to_bytes_kwargs = {'length': _BYTES_PER_TRANSMISSION_CODE}
        self._to_bytes_kwargs.update(self._to_bytes_kwargs_without_length)

        self._shmm_for_results = multiprocessing.shared_memory.SharedMemory(
            create=True,
            size=self._code_count * _BYTES_PER_TRANSMISSION_CODE)

        self._shmm_for_durations = multiprocessing.shared_memory.SharedMemory(
            create=True,
            size=self._code_count * _DURATION_SIZE
        )

        self._lock = self._mp_context.Lock()

        with self._lock:
            self._shmm_for_results.buf[:self._code_count * _BYTES_PER_TRANSMISSION_CODE] = \
                b''.join([TransmissionCode.INITIAL.to_bytes(**self._to_bytes_kwargs)] * self._code_count)

            self._shmm_for_durations.buf[:self._code_count * _DURATION_SIZE] = bytearray(
                [0] * _DURATION_SIZE * self._code_count)

    def get(self) -> typing.Optional[int]:
        with self._lock:
            for i in range(self._code_count):
                start_index = i * _BYTES_PER_TRANSMISSION_CODE
                code = int.from_bytes(
                    self._shmm_for_results.buf[start_index:start_index + _BYTES_PER_TRANSMISSION_CODE],
                    **self._from_bytes_kwargs)
                if code == TransmissionCode.INITIAL:
                    self._shmm_for_results.buf[start_index:start_index + _BYTES_PER_TRANSMISSION_CODE] = \
                        TransmissionCode.STARTED.to_bytes(**self._to_bytes_kwargs)
                    return i

    def put(self, index, transmission_code: TransmissionCode, duration: int):
        assert 0 <= index < self._code_count

        result_index = index * _BYTES_PER_TRANSMISSION_CODE
        duration_index = index * _DURATION_SIZE

        with self._lock:
            assert int.from_bytes(
                self._shmm_for_results.buf[result_index:result_index + _BYTES_PER_TRANSMISSION_CODE],
                **self._from_bytes_kwargs) == TransmissionCode.STARTED, \
                'result input is only allowed after a start code has been written'

            self._shmm_for_results.buf[result_index:result_index + _BYTES_PER_TRANSMISSION_CODE] = \
                transmission_code.to_bytes(**self._to_bytes_kwargs)

            self._shmm_for_durations.buf[duration_index:duration_index + _DURATION_SIZE] = \
                duration.to_bytes(length=_DURATION_SIZE, **self._to_bytes_kwargs_without_length)

    def get_results(self) -> typing.List[typing.Tuple[int, TransmissionCode, int]]:
        results = list()

        with self._lock:
            for i in range(self._code_count):
                result_index = i * _BYTES_PER_TRANSMISSION_CODE
                code = TransmissionCode.from_bytes(
                    self._shmm_for_results.buf[result_index:result_index + _BYTES_PER_TRANSMISSION_CODE],
                    **self._from_bytes_kwargs)

                duration_index = i * _DURATION_SIZE
                duration = int.from_bytes(
                    self._shmm_for_durations.buf[duration_index:duration_index + _DURATION_SIZE],
                    **self._from_bytes_kwargs
                )
                results.append((
                    i,
                    code,
                    duration))

        return results

    def is_finished(self) -> bool:
        results = list()

        with self._lock:
            for i in range(self._code_count):
                result_index = i * _BYTES_PER_TRANSMISSION_CODE
                code = TransmissionCode.from_bytes(
                    self._shmm_for_results.buf[result_index:result_index + _BYTES_PER_TRANSMISSION_CODE],
                    **self._from_bytes_kwargs)
                results.append(
                    code in (TransmissionCode.PASS,
                             TransmissionCode.FAIL,
                             TransmissionCode.SKIPPED,
                             TransmissionCode.ERROR))

        return all(results) and len(results) == self._code_count

    def close(self):
        try:
            self._shmm_for_results.close()
        finally:
            self._shmm_for_durations.close()

    def unlink(self):
        try:
            self._shmm_for_results.unlink()
        finally:
            self._shmm_for_durations.unlink()


class PipeCommunicator(BaseCommunicator):
    _BLOCKING = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._initial_queue = self._mp_context.Queue(maxsize=self._code_count)
        self._result_queue = self._mp_context.Queue(maxsize=self._code_count)

        self._initial_lock = self._mp_context.Lock()
        self._result_lock = self._mp_context.Lock()

        # Acquiring the initial lock might not be necessary here,
        # when the initialization is finished before starting other processes,
        # but since we can not guarantee that behaviour and to be compatible with
        # future changes of process supervision, we still use the lock here
        with self._initial_lock:
            for i in range(self._code_count):
                self._initial_queue.put((i, TransmissionCode.INITIAL), block=self._BLOCKING)

    def get(self) -> typing.Optional[int]:
        with self._initial_lock:
            index, transmission_code = self._initial_queue.get(block=self._BLOCKING)

            if transmission_code == TransmissionCode.STARTED:
                self._initial_queue.put((index, transmission_code), block=self._BLOCKING)
                return None
            elif transmission_code == TransmissionCode.INITIAL:
                self._initial_queue.put((index, TransmissionCode.STARTED), block=self._BLOCKING)
                return index
            else:
                self._initial_queue.put((index, transmission_code), block=self._BLOCKING)
                raise ValueError('Got an unexpected value for transmission code from _initial_queue')

    def put(self, index: int, transmission_code: TransmissionCode, duration: int):
        assert 0 <= index < self._code_count

        with self._result_lock:
            self._result_queue.put((index, transmission_code, duration), block=self._BLOCKING)

    def get_results(self) -> typing.List[typing.Tuple[int, TransmissionCode, int]]:

        unordered_results = list()

        with self._result_lock:

            while len(unordered_results) < self._code_count:
                unordered_results.append(self._result_queue.get(block=self._BLOCKING))

            for result in unordered_results:
                self._result_queue.put(result, block=self._BLOCKING)

        assert len({i for i, _, __ in unordered_results}) == len(unordered_results) == self._code_count

        results = list()
        for index, result, duration in sorted(unordered_results, key=lambda e: e[0]):
            results.append((index, result, duration))

        return results

    def is_finished(self):
        with self._result_lock:
            # TODO: (consider) replacement, because it is unreliable according to its documentation
            return self._result_queue.full()

    def close(self):
        pass  # nothing to do here, because emptying and closing the queues once (in unlink) is enough

    def unlink(self):
        with self._initial_lock:
            for _ in range(self._code_count):
                self._initial_queue.get(block=self._BLOCKING)

            self._initial_queue.close()

        with self._result_lock:
            for _ in range(self._code_count):
                self._result_queue.get(block=self._BLOCKING)

            self._result_queue.close()



"""
This file is part of mpunittest, a parallel unittest runner.
Copyright (C) 2023 Maximilian Pilz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import contextlib
import os
import pathlib
import sys
import typing


class RedirectionContextManagerFactory(typing.Protocol):
    def __call__(self, file: pathlib.Path) -> typing.ContextManager[typing.TextIO]: ...


@contextlib.contextmanager
def redirect_stderr_to_file(file: pathlib.Path) -> typing.Generator:
    """
    Redirect sys.stderr to the given file.

    :param file: target of redirection
    :return: a generator, which can be converted to a respective contextmanager
    """
    assert file.is_file() or not file.exists()

    with open(file, 'w') as f:
        with redirect_fd(f.fileno(), sys.stderr.fileno()):
            try:
                yield
            finally:
                sys.stderr.flush()


@contextlib.contextmanager
def redirect_stdout_to_file(file: pathlib.Path) -> typing.Generator:
    """
    Redirect sys.stdout to the given file.

    :param file: target of redirection
    :return: a generator, which can be converted to a respective contextmanager
    """
    assert file.is_file() or not file.exists()

    with open(file, 'w') as f:
        with redirect_fd(f.fileno(), sys.stdout.fileno()):
            try:
                yield
            finally:
                sys.stdout.flush()


@contextlib.contextmanager
def redirect_fd(dst_fd: int, src_fd: int) -> typing.Generator:
    """
    Do the following:
        1. Allocate a new file descriptor and make it refer to the file description of src_fd
            i.e. create a backup of src_fd.
        2. Adjust the file descriptor src_fd so that it refers to the same open file description as dst_fd.
        3. Execute code of context via yield.
        4. Readjust the file descriptor src_fd so that it refers to the same open file description as it
            did at the beginning of the function i.e. reset it to its original state.
        5. Close the new file descriptor that was used as a backup.

    :param dst_fd: file descriptor that refers to the file description to be used for adjusting src_fd
    :param src_fd: file descriptor to be adjusted
    :return: a generator, which can be converted to a respective contextmanager
    """
    old_src_fd: int = os.dup(src_fd)  # backup

    try:
        os.dup2(dst_fd, src_fd)  # redirection
        try:
            yield  # code in the context of the context manager will be run here
        finally:
            os.dup2(old_src_fd, src_fd)  # reset
    finally:
        os.close(old_src_fd)  # since old_src_fd was only needed as a backup it can be closed now



"""
This file is part of mpunittest, a parallel unittest runner.
Copyright (C) 2023 Maximilian Pilz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import os
import pathlib
import tempfile
import unittest
import unittest.mock

import mpunittest.result
import mpunittest.runner
import mpunittest.tests.dummy.dirs


class TestRunner(unittest.TestCase):

    def test_flatten(self):

        test_case_1 = unittest.mock.Mock()
        test_case_2 = unittest.mock.Mock()
        test_case_3 = unittest.mock.Mock()

        test_suite = unittest.TestSuite()
        test_suite.addTest(test_case_1)
        test_suite.addTest(test_case_2)

        self.assertEqual(
            (test_case_1, test_case_2),
            tuple(mpunittest.runner.MergingRunner.flatten(test_suite))
        )

        self.assertEqual(
            (test_case_3,),
            tuple(mpunittest.runner.MergingRunner.flatten(test_case_3))
        )

    def test_time_conversion(self):

        self.assertEqual('0.000000001 ' + mpunittest.runner.TimeUnit.SECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             1,
                             mpunittest.runner.TimeUnit.SECONDS)
                         )

        self.assertEqual('0.1 ' + mpunittest.runner.TimeUnit.SECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             int(10 ** 8),
                             mpunittest.runner.TimeUnit.SECONDS)
                         )

        self.assertEqual('1 ' + mpunittest.runner.TimeUnit.SECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             int(10 ** 9),
                             mpunittest.runner.TimeUnit.SECONDS)
                         )

        self.assertEqual('11.67 ' + mpunittest.runner.TimeUnit.SECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             1167 * int(10 ** 7),
                             mpunittest.runner.TimeUnit.SECONDS)
                         )

        self.assertEqual('1.55 ' + mpunittest.runner.TimeUnit.MILLISECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             1550000,
                             mpunittest.runner.TimeUnit.MILLISECONDS)
                         )

        self.assertEqual('1 ' + mpunittest.runner.TimeUnit.MILLISECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             1000000,
                             mpunittest.runner.TimeUnit.MILLISECONDS)
                         )

        self.assertEqual('0.5 ' + mpunittest.runner.TimeUnit.MILLISECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             500000,
                             mpunittest.runner.TimeUnit.MILLISECONDS)
                         )

        self.assertEqual('0.5 ' + mpunittest.runner.TimeUnit.MICROSECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             500,
                             mpunittest.runner.TimeUnit.MICROSECONDS)
                         )

        self.assertEqual('5 ' + mpunittest.runner.TimeUnit.MICROSECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             5000,
                             mpunittest.runner.TimeUnit.MICROSECONDS)
                         )

        self.assertEqual('50 ' + mpunittest.runner.TimeUnit.MICROSECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             50000,
                             mpunittest.runner.TimeUnit.MICROSECONDS)
                         )

        self.assertEqual('500 ' + mpunittest.runner.TimeUnit.NANOSECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             500,
                             mpunittest.runner.TimeUnit.NANOSECONDS)
                         )

        self.assertEqual('5000000000 ' + mpunittest.runner.TimeUnit.NANOSECONDS.name.lower(),
                         mpunittest.runner.ns_to_time_unit(
                             5000000000,
                             mpunittest.runner.TimeUnit.NANOSECONDS)
                         )

    def test_discover_and_run(self):

        for ipc_method in mpunittest.runner.IPCMethod:
            with self.subTest(msg=f'test with "{ipc_method}"'):
                with tempfile.TemporaryDirectory() as temp_dir:
                    temp_path = pathlib.Path(temp_dir)

                    results = mpunittest.runner.MergingRunner(ipc_method=ipc_method).discover_and_run(
                        start_dir=pathlib.Path(mpunittest.tests.dummy.dirs.__file__).parent.resolve(),
                        pattern="*test.py",
                        html_result_assets=mpunittest.runner.HtmlResultAssets(
                            document_title=None,  # this will cause usage of the default title
                            document_file_name=None,  # this will cause usage of the default file name
                            result_path=temp_path
                        )
                    )

                    result_mapping = dict()

                    for test_id, simple_result, duration, _ in results:
                        self.assertNotIn(test_id, result_mapping)
                        result_mapping[test_id] = simple_result

                    self.assertEqual(7, len(set(result_mapping.keys())))
                    self.assertCountEqual([mpunittest.result.SimpleResult.ERROR,
                                           mpunittest.result.SimpleResult.ERROR,
                                           mpunittest.result.SimpleResult.PASS,
                                           mpunittest.result.SimpleResult.PASS,
                                           mpunittest.result.SimpleResult.SKIPPED,
                                           mpunittest.result.SimpleResult.PASS,
                                           mpunittest.result.SimpleResult.FAIL],
                                          result_mapping.values())

                    self.assertNotIn('<!--', temp_path.joinpath('test_result.html').read_text())


if __name__ == '__main__':
    unittest.main()



"""
This file is part of mpunittest, a parallel unittest runner.
Copyright (C) 2023 Maximilian Pilz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import collections
import contextlib
import dataclasses
import enum
import multiprocessing
import multiprocessing.connection
import multiprocessing.shared_memory
import os
import pathlib
import re
import tempfile
import time
import typing
import unittest
import uuid

import mpunittest.comm
import mpunittest.html
import mpunittest.logging
import mpunittest.result
import mpunittest.streamctx

_tr_template = \
    """
<tr id="{html_id_string}">
    <td>{test_id}</td>
    <td>{duration}</td>
    <td>
        <a href="{log_file}" class="button {extra_class}"><span>{text}</span></a>
    </td>
</tr>
    """

HtmlResultAssets = collections.namedtuple('HtmlResultAssets', ('document_title', 'document_file_name', 'result_path'))


class TempDirOrDummyContextManagerFactory(typing.Protocol):
    def __call__(self) -> typing.ContextManager[str]: ...


@dataclasses.dataclass
class LogFileParameters:
    log_file_names: typing.List[typing.Optional[pathlib.Path]]

    temp_dir_ctx: TempDirOrDummyContextManagerFactory
    stderr_redirect_ctx: mpunittest.streamctx.RedirectionContextManagerFactory
    stdout_redirect_ctx: mpunittest.streamctx.RedirectionContextManagerFactory


class TimeUnit(int, enum.Enum):
    """
    Time units given in nanoseconds.
    """
    NANOSECONDS = 0
    MICROSECONDS = 3
    MILLISECONDS = 6
    SECONDS = 9


def ns_to_time_unit(time_spent: int, time_unit: TimeUnit) -> str:
    time_digits = list()
    time_digits.append(time_spent // (10 ** time_unit))
    next_digits = time_spent % (10 ** time_unit)
    if next_digits:
        time_digits.append('0' * (time_unit - len(str(next_digits))) + str(next_digits).rstrip('0'))
    return '.'.join(map(str, time_digits)) + ' ' + time_unit.name.lower()


class IPCMethod(enum.Enum):
    SHARED_MEMORY = 0
    PIPE_BASED_QUEUE = 1


class MergingRunner:
    """
    An unittest runner for parallel unittest execution that (optionally) can merge results into a file.
    """
    _logger = mpunittest.logging.logger.getChild('runner')

    _regex = re.compile('(?s)<!--.*-->\n')

    def __init__(self,
                 process_count: int = 2,
                 mp_context: multiprocessing.context.BaseContext = None,
                 daemons: bool = True,
                 result_class=mpunittest.result.SubProcessResult,
                 ipc_method: IPCMethod = IPCMethod.SHARED_MEMORY):
        """
        :param process_count: amount of process to start for delegating unittest execution to them
        :param mp_context: multiprocessing context to use for starting the processes e.g. the 'spawn' context
        :param daemons: will be used as daemon flag for process creation
        :param result_class: type to instantiate for saving unittest results
        """
        assert process_count > 0
        self._process_count = process_count

        if mp_context is None:
            self._mp_context: multiprocessing.context.BaseContext = multiprocessing.get_context('spawn')
        else:
            self._mp_context: multiprocessing.context.BaseContext = mp_context

        self._result_class = result_class
        self._daemons = daemons

        self._ipc_method = ipc_method

        if self._ipc_method == IPCMethod.SHARED_MEMORY:
            self._logger.warning(f'{self.__class__.__name__} instance is configured to use '
                                 f'the {IPCMethod.__name__} "{self._ipc_method}", which is an '
                                 f'experimental feature, there may be memory leaks outside of '
                                 f'the memory space of processes. If you do not want this, '
                                 f'you can specify another {IPCMethod.__name__} such as '
                                 f'{IPCMethod.PIPE_BASED_QUEUE} to be used')

    @staticmethod
    def flatten(test_obj: typing.Union[unittest.TestSuite, unittest.TestCase]) -> typing.Generator:
        """
        Generate individual test case values out of test suites and do an identity mapping in case of a test case.

        :param test_obj: test suite or test case, for test suites all test cases will be extracted from it
        :return: a generator for test cases
        """
        try:
            for i in test_obj:
                for e in MergingRunner.flatten(i):
                    yield e
        except TypeError:
            yield test_obj

    @staticmethod
    def _discover_ids(*args, **kwargs) -> list:

        loader = unittest.loader.TestLoader()
        test_generator = MergingRunner.flatten(loader.discover(*args, **kwargs))

        return [test.id() for test in test_generator]

    @staticmethod
    def _discover_id_to_test_mapping(*args, **kwargs) -> dict:

        loader = unittest.loader.TestLoader()
        test_generator = MergingRunner.flatten(loader.discover(*args, **kwargs))

        return {test.id(): test for test in test_generator}

    @staticmethod
    def _get_log_file_parameters(result_path, result_file_count: int):

        if result_path:
            log_file_names = [result_path.joinpath('test' + uuid.uuid4().hex + '.log')
                              for _ in range(result_file_count)]
            assert len(set(log_file_names)) == len(log_file_names)

            temp_dir_ctx = tempfile.TemporaryDirectory
            stderr_redirect_ctx = mpunittest.streamctx.redirect_stderr_to_file
            stdout_redirect_ctx = mpunittest.streamctx.redirect_stdout_to_file

        else:
            log_file_names = [None] * result_file_count

            temp_dir_ctx = MergingRunner.dummy_context
            stderr_redirect_ctx = MergingRunner.dummy_context
            stdout_redirect_ctx = MergingRunner.dummy_context

        return LogFileParameters(log_file_names=log_file_names,
                                 temp_dir_ctx=temp_dir_ctx,
                                 stderr_redirect_ctx=stderr_redirect_ctx,
                                 stdout_redirect_ctx=stdout_redirect_ctx)

    def discover_and_run(
            self,
            start_dir: pathlib.Path,
            pattern: str = 'test*.py',
            top_level_dir: str = None,
            html_result_assets: HtmlResultAssets = None,
            time_unit: TimeUnit = TimeUnit.SECONDS
    ) -> typing.List[typing.Tuple[str, mpunittest.result.SimpleResult, int, pathlib.Path]]:
        """
        Discover test cases in modules matching the given pattern in the given directory.

        :param start_dir: directory to search in
        :param pattern: pattern to check modules against
        :param top_level_dir: same as top_level_dir parameter of unittest.loader.discover
        :param html_result_assets: if given an HTML file containing the results will be generated
        according to the parameters in the given asset or with default parameters in case all are None,
        otherwise no files are generated
        :param time_unit: unit for displaying the time spent
        :return: TODO
        """
        start = time.monotonic_ns()

        if html_result_assets:
            if html_result_assets.result_path:
                result_path = html_result_assets.result_path
            else:
                result_path = pathlib.Path(f'testruns') / pathlib.Path(f'testrun{uuid.uuid4().hex}')
            result_path.mkdir(parents=True, exist_ok=True)

            if html_result_assets.document_title:
                doc_title = html_result_assets.document_title
            else:
                doc_title = 'Unittest results'

            if html_result_assets.document_file_name:
                html_file_name = html_result_assets.document_file_name
            else:
                html_file_name = 'test_result'
        else:
            result_path = None
            doc_title = None
            html_file_name = None

        self._logger.debug('primary runner process will start discovery in directory "%s" '
                           'with pattern "%s" and top level directory "%s"',
                           start_dir, pattern, top_level_dir)
        test_ids = MergingRunner._discover_ids(start_dir=start_dir,
                                               pattern=pattern,
                                               top_level_dir=top_level_dir)
        self._logger.info('primary runner process finished discovery')
        test_id_count = len(test_ids)
        self._logger.info('primary runner process discovered %i test(s) in "%s"',
                          test_id_count,
                          start_dir.as_uri())

        log_file_parameters: LogFileParameters = self._get_log_file_parameters(
            result_path=result_path,
            result_file_count=test_id_count
        )

        if self._ipc_method == IPCMethod.SHARED_MEMORY:  # TODO: change type of _ipc_method to call it here
            communicator = mpunittest.comm.SharedMemoryCommunicator(
                code_count=test_id_count, mp_context=self._mp_context
            )
        elif self._ipc_method == IPCMethod.PIPE_BASED_QUEUE:
            communicator = mpunittest.comm.PipeCommunicator(
                code_count=test_id_count, mp_context=self._mp_context)
        else:
            raise NotImplementedError

        try:
            child_processes = list()

            # TODO: consider using test_ids without enumeration
            enumerated_test_ids = tuple(enumerate(test_ids))  # TODO: evaluate whether this is really necessary

            for _ in range(self._process_count):
                child_process = self._mp_context.Process(daemon=self._daemons,
                                                         target=self.process_target,
                                                         args=(enumerated_test_ids,
                                                               communicator,
                                                               start_dir,
                                                               pattern,
                                                               top_level_dir,
                                                               self._result_class,
                                                               log_file_parameters
                                                               ,))

                self._logger.debug('will start process with name "%s"', child_process.name)
                child_process.start()
                self._logger.info('started process with name "%s"', child_process.name)

                child_processes.append(child_process)

            assert len(child_processes) == self._process_count

            self._logger.debug('will wait for child processes to finish')
            assert child_processes
            for child_process in child_processes:
                child_process.join()
                exit_code = child_process.exitcode

                # since no timeout was provided exit_code is not allowed to be None
                assert exit_code is not None, 'exitcode was None even though no timeout was provided to join'

                if not exit_code == 0:
                    self._logger.error('process with name "%s" exited with code "%i"',
                                       child_process.name,
                                       exit_code)
                    raise RuntimeError(f'Expected exitcode of {child_process.name} to be 0, '
                                       f'but it is {exit_code} instead.')

            if not communicator.is_finished():
                raise RuntimeError('All subprocesses have been joined, but not all results are known.')

            test_results = communicator.get_results()
        finally:
            communicator.close()
            communicator.unlink()

        end = time.monotonic_ns()
        total_time_spent_ns = end - start  # TODO: maybe add up the subprocess times instead

        assert len(test_results) == test_id_count

        converted_results = [(test_ids[result_index],
                              mpunittest.result.TransmissionCodeToSimpleResult[result_transmission_code],
                              result_duration,
                              log_file_parameters.log_file_names[result_index])
                             for result_index, result_transmission_code, result_duration in test_results]

        assert all(c in mpunittest.result.SimpleResult for _, c, __, ___ in converted_results)

        if html_result_assets:
            self._logger.debug('will generate html file in %s', result_path)
            self._generate_html(test_results=converted_results,
                                result_path=result_path,
                                total_time_spent_ns=total_time_spent_ns,
                                doc_title=doc_title,
                                html_file_name=html_file_name,
                                time_unit=time_unit)
            self._logger.info('generated html file in %s', result_path)

        return converted_results

    @staticmethod
    def _generate_html(test_results: typing.List[typing.Tuple[str, mpunittest.result.SimpleResult, int, pathlib.Path]],
                       result_path: pathlib.Path,
                       total_time_spent_ns: int,
                       doc_title: str,
                       html_file_name: str,
                       time_unit: TimeUnit):
        with open(
                pathlib.Path(mpunittest.html.__file__).parent.joinpath('result.html'), 'r'
        ) as html_template:
            template_data = html_template.read()

        matches = MergingRunner._regex.findall(template_data)
        assert matches
        assert len(matches) == 1
        for match in matches:
            template_data = template_data.replace(match, '')

        table_row_data = list()
        for test_id, simple_result, duration, log_file in test_results:
            assert log_file.is_file()

            table_row_data.append(
                (test_id,
                 duration,
                 log_file.name,
                 simple_result)
            )
        table_row_data = [(i, *d) for i, d in enumerate(table_row_data)]

        table_rows_string = str()

        pass_count = 0
        fail_count = 0
        skip_count = 0
        error_count = 0

        for index, test_id, time_spent, log_file_name, extra_class in table_row_data:
            table_rows_string += _tr_template.format(
                html_id_string=str(index),
                test_id=test_id,
                duration=ns_to_time_unit(time_spent=time_spent, time_unit=time_unit),
                log_file=log_file_name,
                extra_class=extra_class.name.lower(),
                text=extra_class.name.upper()
            )

            if extra_class == mpunittest.result.SimpleResult.PASS:
                pass_count += 1
            elif extra_class == mpunittest.result.SimpleResult.FAIL:
                fail_count += 1
            elif extra_class == mpunittest.result.SimpleResult.SKIPPED:
                skip_count += 1
            elif extra_class == mpunittest.result.SimpleResult.ERROR:
                error_count += 1
            else:
                # TODO: also call logger here
                raise ValueError(f'Unexpected value for extra_class: '
                                 f'"{extra_class}" of type "{type(extra_class)}"')

        total_count = sum((pass_count, fail_count, skip_count, error_count))

        original_order = [e[0] for e in table_row_data]

        # sort by test_id
        sorted_by_name_asc = [e[0] for e in sorted(
            table_row_data, key=lambda e: e[1])]
        sorted_by_name_desc = [e[0] for e in sorted(
            table_row_data, key=lambda e: e[1], reverse=True)]

        # sort by time_spent
        sorted_by_time_asc = [e[0] for e in sorted(
            table_row_data, key=lambda e: e[2])]
        sorted_by_time_desc = [e[0] for e in sorted(
            table_row_data, key=lambda e: e[2], reverse=True)]

        with open(result_path.joinpath(f'{html_file_name}.html'), 'w') as final_html_file:
            final_html_data = template_data.format(
                title=doc_title,
                table_title=doc_title,
                time=f'{ns_to_time_unit(time_spent=total_time_spent_ns, time_unit=time_unit)}',
                total_count=total_count,
                pass_count=pass_count,
                fail_count=fail_count,
                skip_count=skip_count,
                error_count=error_count,
                table_rows=table_rows_string,
                ordered_by_name=','.join(
                    map(str,
                        [
                            original_order,
                            sorted_by_name_asc,
                            sorted_by_name_desc
                        ])),
                ordered_by_time=','.join(
                    map(str,
                        [
                            original_order,
                            sorted_by_time_asc,
                            sorted_by_time_desc
                        ]))
            )

            final_html_file.write(final_html_data)

    @staticmethod
    def process_target(
            enumerated_test_ids,
            communicator,
            start_dir: pathlib.Path,
            pattern: str,
            top_level_dir: str,
            result_class: typing.Type[mpunittest.result.SubProcessResult],
            log_file_parameters: LogFileParameters
    ) -> None:
        """
        Run inside of child processes and execute test cases with the ids that are send to the
        respective child process.
        Also send back the result of the execution and optionally write stderr and stdout in a file
        together with the result.

        :param enumerated_test_ids: TODO
        :param start_dir: directory to search for test cases to eventually load them and execute them
        if requested
        :param pattern: pattern to check modules against
        :param top_level_dir: same as top_level_dir parameter of unittest.loader.discover
        :param result_class: type to instantiate for saving unittest results
        :param result_path: path to generate result files in, can be None to not generate any files
        :return: None
        """
        try:
            index_to_test_id = {i: e for i, e in enumerated_test_ids}

            temp_dir_ctx = log_file_parameters.temp_dir_ctx
            stderr_redirect_ctx = log_file_parameters.stderr_redirect_ctx
            stdout_redirect_ctx = log_file_parameters.stdout_redirect_ctx

            id_to_test_mapping = MergingRunner._discover_id_to_test_mapping(start_dir, pattern, top_level_dir)

            time_str = str(time.time()).replace('.', '_')  # TODO: remove when no longer needed
            filename_postfix = f'pid{os.getpid()}_t{time_str}'
            stderr_filename = 'stderr' + filename_postfix
            stdout_filename = 'stdout' + filename_postfix

            while True:

                test_id_index = communicator.get()
                if test_id_index is None:
                    break
                test_id = index_to_test_id[test_id_index]
                log_file_name = log_file_parameters.log_file_names[test_id_index]

                result = result_class()  # TODO: add parameters here such as stream and verbosity

                with temp_dir_ctx() as temp_dir:  # TODO: move out of loop

                    if temp_dir:
                        temp_path = pathlib.Path(temp_dir).resolve()
                        final_stderr_filename = temp_path.joinpath(stderr_filename)
                        final_stdout_filename = temp_path.joinpath(stdout_filename)
                    else:
                        final_stderr_filename = None
                        final_stdout_filename = None

                    with \
                            stderr_redirect_ctx(final_stderr_filename), \
                            stdout_redirect_ctx(final_stdout_filename):

                        result.startTestRun()
                        # execute test here
                        id_to_test_mapping[test_id](result)
                        result.stopTestRun()

                    transmission_code, str_output = result.overall_result

                    if log_file_name:
                        with \
                                open(log_file_name, 'wb') as merged, \
                                open(final_stderr_filename, 'rb') as stderr_file, \
                                open(final_stdout_filename, 'rb') as stdout_file:
                            merged.write(f'Log for "{test_id}" '
                                         f'run at {time_str}: \n\n'.encode('utf8'))

                            merged.write(('#' * 10 + ' ' * 2 + 'stderr' + ' ' * 2 + '#' * 10 + '\n\n').encode('utf8'))
                            MergingRunner.copy_content_with_limited_buffer(stderr_file, merged)
                            merged.write(('\n' + '#' * 30 + '\n\n').encode('utf8'))
                            merged.write(('#' * 10 + ' ' * 2 + 'stdout' + ' ' * 2 + '#' * 10 + '\n\n').encode('utf8'))
                            MergingRunner.copy_content_with_limited_buffer(stdout_file, merged)
                            merged.write(('\n' + '#' * 30 + '\n\n').encode('utf8'))
                            merged.write(f'Overall result: {transmission_code.name}\n'.encode('utf8'))
                            merged.write(f'\n{str_output}\n'.encode('utf8'))

                communicator.put(index=test_id_index,
                                 transmission_code=transmission_code,
                                 duration=result.time_spent)
        finally:
            communicator.close()

    @staticmethod
    def copy_content_with_limited_buffer(
            src: typing.IO,
            dst: typing.IO,
            buffer_size: int = 1000
    ) -> None:
        """
        Copy data from src to dst.

        :param src: io object to read from
        :param dst: io object to write to i.e. to copy to
        :param buffer_size: maximum amount of unit defined by the given io objects to read per iteration
        :return: None
        """
        while True:
            data = src.read(buffer_size)
            if not data:
                break
            dst.write(data)

    @staticmethod
    @contextlib.contextmanager
    def dummy_context(*_, **__) -> typing.Generator:
        """
        Do not do anything additionally to the code in the context.
        This is supposed to be a syntactical replacement for other contexts in case that the other contexts shall not
        be used.

        :return: generator that can be converted into context manager that does effectively nothing
        """
        yield



"""
This file is part of mpunittest, a parallel unittest runner.
Copyright (C) 2023 Maximilian Pilz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import argparse
import logging
import pathlib

import mpunittest.runner

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog='mpunittest.run',
        description='mpunittest.run executes unittests in the given '
                    'directory with the given amount of processes',
        epilog=f'mpunittest in version {mpunittest.VERSION}')

    parser.add_argument('directory')
    parser.add_argument('-p', '--pattern', required=False)
    parser.add_argument('-t', '--top_level_dir', required=False)
    parser.add_argument('-c', '--count', help='amount of processes to use', required=False)
    parser.add_argument('-d', '--daemon',
                        help='whether child processes shall be daemon processes',
                        action='store_true', required=False)
    parser.add_argument('-dt', '--document_title', required=False)
    parser.add_argument('-dfn', '--document_file_name', required=False)
    parser.add_argument('-o', '--output_dir', required=False)
    parser.add_argument('-l', '--log_level', required=False, default='INFO',
                        help='CRITICAL, FATAL, ERROR, WARNING, INFO, DEBUG')
    parser.add_argument('-da', '--default_asset', action='store_true', required=False)

    args = parser.parse_args()

    handler = logging.StreamHandler()

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    handler.setFormatter(formatter)
    handler.setLevel(getattr(logging, args.log_level))
    mpunittest.logging.logger.addHandler(handler)

    if args.count:
        merging_runner = mpunittest.runner.MergingRunner(process_count=int(args.count),
                                                         daemons=args.daemon)
    else:
        merging_runner = mpunittest.runner.MergingRunner(daemons=args.daemon)

    runner_args = {
        'start_dir': pathlib.Path(args.directory).resolve()
    }
    if args.pattern:
        runner_args['pattern'] = args.pattern
    if args.top_level_dir:
        runner_args['top_level_dir'] = args.top_level_dir

    asset_args = dict()

    if args.document_title:
        asset_args['document_title'] = args.document_title
    if args.document_file_name:
        asset_args['document_file_name'] = args.document_file_name
    if args.output_dir:
        asset_args['result_path'] = args.output_dir

    if args.default_asset and asset_args:
        raise RuntimeError('result asset parameters are inconsistent')

    if asset_args:
        runner_args['html_result_assets'] = mpunittest.runner.HtmlResultAssets(
            **asset_args
        )
    if args.default_asset:
        runner_args['html_result_assets'] = mpunittest.runner.HtmlResultAssets(None, None, None)

    result = merging_runner.discover_and_run(
        **runner_args
    )

    mpunittest.logging.logger.getChild('tests.real.manual').info(
        'run finished with return value %s', result)



"""
This file is part of mpunittest, a parallel unittest runner.
Copyright (C) 2023 Maximilian Pilz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import enum
import time
import typing
import unittest

import mpunittest.comm


class SimpleResult(int, enum.Enum):  # TODO: change to enum.StrEnum in CPython 3.11
    PASS = 1
    FAIL = 2
    SKIPPED = 3
    ERROR = 4


TransmissionCodeToSimpleResult = {
    mpunittest.comm.TransmissionCode.PASS: SimpleResult.PASS,
    mpunittest.comm.TransmissionCode.FAIL: SimpleResult.FAIL,
    mpunittest.comm.TransmissionCode.SKIPPED: SimpleResult.SKIPPED,
    mpunittest.comm.TransmissionCode.ERROR: SimpleResult.ERROR
}


class SubProcessResult(unittest.TestResult):

    @property
    def time_spent(self):
        assert self._time_spent is not None
        assert self._time_spent >= 0

        return self._time_spent

    @property
    def overall_result(self) -> typing.Tuple[
        typing.Literal[
            mpunittest.comm.TransmissionCode.PASS,
            mpunittest.comm.TransmissionCode.FAIL,
            mpunittest.comm.TransmissionCode.SKIPPED,
            mpunittest.comm.TransmissionCode.ERROR], str]:
        assert self.testsRun > 0
        assert self._time_spent >= 0
        assert len(self.skipped) <= self.testsRun

        if self.wasSuccessful() and (len(self.skipped) < self.testsRun):
            assert self.successes
            transmission_code = mpunittest.comm.TransmissionCode.PASS
        elif self.wasSuccessful() and (len(self.skipped) == self.testsRun):
            assert self.skipped
            transmission_code = mpunittest.comm.TransmissionCode.SKIPPED
        elif self.errors:
            assert self.errors
            transmission_code = mpunittest.comm.TransmissionCode.ERROR
        elif self.failures or self.unexpectedSuccesses:
            assert self.failures
            transmission_code = mpunittest.comm.TransmissionCode.FAIL
        else:
            raise NotImplementedError  # TODO: add parameter

        str_output = '\n'

        if self.errors:
            for test, err_str in self.errors:
                str_output += f'The test {test.id()} had an error.\n'
                if err_str:
                    str_output += f'Error:\n{err_str}\n'
                str_output += '\n'
        if self.failures:
            for test, err_str in self.failures:
                str_output += f'The test {test.id()} had a failure.\n'
                if err_str:
                    str_output += f'Failure:\n{err_str}\n'
                str_output += '\n'
        if self.unexpectedSuccesses:
            for test in self.unexpectedSuccesses:
                str_output += f'The test {test.id()} had a failure due to an unexpected success.\n\n'
        if self.skipped:
            for test, reason in self.skipped:
                str_output += f'The test {test.id()} was skipped.\n'
                str_output += f'Reason: {reason}\n\n'
        if self.successes:
            for test in self.successes:
                str_output += f'The test {test.id()} was successful.\n\n'

        return transmission_code, str_output

    def wasSuccessful(self) -> bool:
        return (super().wasSuccessful() and
                (self.successes or
                 self.expectedFailures or
                 self.skipped) and
                self.testsRun > 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        assert self.buffer is False

        self._time_spent: typing.Optional[int] = None
        assert not hasattr(self, 'successes')
        self.successes: typing.List[unittest.case.TestCase] = list()

    def startTestRun(self) -> None:
        assert self.buffer is False

        super().startTestRun()

        assert self._time_spent is None
        self._time_spent = -time.monotonic_ns()

    def stopTestRun(self) -> None:
        super().stopTestRun()

        self._time_spent += time.monotonic_ns()
        assert self._time_spent >= 0

    def startTest(self, test: unittest.case.TestCase) -> None:
        assert self.buffer is False

        super().startTest(test)

    @unittest.result.failfast
    def addError(self, *args, **kwargs) -> None:
        error_count = len(self.errors)

        super().addError(*args, **kwargs)

        assert len(self.errors) > error_count

        assert self.errors[-1]
        # print(self.errors[-1][1], file=sys.stderr, flush=True)

    @unittest.result.failfast
    def addFailure(self, *args, **kwargs) -> None:
        failure_count = len(self.failures)

        super().addFailure(*args, **kwargs)

        assert len(self.failures) > failure_count

        assert self.failures[-1]
        # print(self.failures[-1][1], file=sys.stderr, flush=True)

    def addSubTest(self, *args, **kwargs) -> None:
        error_count = len(self.errors)
        failure_count = len(self.failures)

        super().addSubTest(*args, **kwargs)

        assert (len(self.errors) > error_count) or (len(self.failures) > failure_count)

        # if len(self.errors) > error_count:
        #     print(self.errors[-1][1], file=sys.stderr, flush=True)
        # else:
        #     print(self.failures[-1][1], file=sys.stderr, flush=True)

    def addSuccess(self, test: unittest.case.TestCase) -> None:
        super().addSuccess(test)

        self.successes.append(test)
